@@ -3,7 +3,7 @@ Python wrapper for GitLab API v3
 """
 
 __title__ = "gitlab3"
-__version__ = "0.2.1"
+__version__ = "0.3.0"
 __author__ = "Alex Van't Hof"
 __license__ = "LGPLv3"
 __copyright__ = "Copyright 2013 Alex Van't Hof"
@@ -12,12 +12,13 @@ __copyright__ = "Copyright 2013 Alex Van't Hof"
 import json
 import re
 import requests
+from datetime import tzinfo, timedelta, datetime
 from math import ceil
 
-import exceptions
-from _api_definition import GitLab as _GitLabAPIDefinition
-from _api_definition import _LIST, _GET, _ADD, _EDIT, _DELETE, _HTTP_GET, \
-                            _HTTP_PUT, _HTTP_POST, _HTTP_DELETE
+from . import exceptions
+from ._api_definition import GitLab as _GitLabAPIDefinition
+from ._api_definition import _LIST, _GET, _ADD, _EDIT, _DELETE, \
+                             _HTTP_GET, _HTTP_PUT, _HTTP_POST, _HTTP_DELETE
 
 # Access level constants
 ACCESS_LEVEL_GUEST = 10
@@ -28,6 +29,33 @@ ACCESS_LEVEL_MASTER = 40
 # Maximum 'per_page' value allowed by GitLab when listing
 _MAX_PER_PAGE = 100
 
+
+def _query_list(api_cls, parent, data):
+    """Helper for find and list functions. Queries GitLab for an entire
+       listing of objects '_MAX_PER_PAGE' objects at a time.
+    """
+    data['per_page'] = _MAX_PER_PAGE
+    page = 0
+    last_objs = None
+    while True:
+        data['page'] = page
+        objs = parent._get(api_cls._uq_url, data=data)
+        # GitLab doesn't always return empty list at end, may repeat last...
+        if not objs or str(objs) == last_objs:
+            # Some listings start with page 0, others with 1...
+            # (in the latter case, page 0 will be the same as page 1
+            if page == 1:  # If page 0 == page 1, try page 2
+                page += 1
+                continue
+            break
+        # "hash" the list as a string for comparison to prevent deep copy
+        # (would be modified when converting date strings to datetime objects)
+        last_objs = str(objs)
+        page += 1
+        for obj in objs:
+            yield api_cls(parent, obj)
+
+
 def _add_list_fn(api, api_definition, parent):
     """Create a <PARENT_API>.<name>s() function"""
     def fn(limit=None, page=None, per_page=None, **data):
@@ -59,17 +87,8 @@ def _add_list_fn(api, api_definition, parent):
                 for obj in objs:
                     ret.append(api(parent, obj))
         else:  # Obtain full list
-            data['per_page'] = _MAX_PER_PAGE
-            data['page'] = 1
-            last_objs = None  # GitLab doesn't always return empty list at end
-            while True:
-                objs = parent._get(api._uq_url, data=data)
-                if not objs or objs == last_objs:  # ouch...pricey
-                    break
-                for obj in objs:
-                    ret.append(api(parent, obj))
-                last_objs = objs
-                data['page'] += 1
+            for api_obj in _query_list(api, parent, data):
+                ret.append(api_obj)
         return ret
     setattr(parent, api_definition.plural_name(), fn)
     return fn
@@ -84,7 +103,7 @@ def _find_matches(objects, kwargs, find_all):
         match = True
         # Match all supplied parameters
         for param, val in kwargs.iteritems():
-            if not getattr(obj, param, None) == val:
+            if not getattr(obj, param) == val:
                 match = False
                 break
         if match:
@@ -92,10 +111,12 @@ def _find_matches(objects, kwargs, find_all):
                 ret.append(obj)
             else:
                 return obj
+    if not find_all:
+        return None
     return ret
 
 
-def _add_find_fn(name, list_fn, parent):
+def _add_find_fn(api, name, parent):
     """Create a <PARENT_API>.find_<name>() function"""
     def fn(**kwargs):
         if not kwargs:
@@ -111,23 +132,10 @@ def _add_find_fn(name, list_fn, parent):
             del kwargs['find_all']
         except KeyError:
             find_all = False
+        if not objects:
+            objects = _query_list(api, parent, {})
 
-        if objects:  # A cached list was supplied
-            return _find_matches(objects, kwargs, find_all)
-        elif find_all:  # Have to search entire list of objects
-            return _find_matches(list_fn(), kwargs, find_all)
-        # Optimize slightly and don't obtain whole object list if not necessary
-        ret = []
-        page = 1
-        while True:
-            objects = list_fn(page=page, per_page=_MAX_PER_PAGE)
-            if not objects:
-                break
-            match = _find_matches(objects, kwargs, find_all)
-            if match:
-                return match
-            page += 1
-        return ret
+        return _find_matches(objects, kwargs, find_all)
     setattr(parent, 'find_' + name, fn)
 
 
@@ -147,13 +155,24 @@ def _add_create_fn(api, api_definition, parent):
     """Create a <PARENT_API>.add_<name>() function"""
     fn_name = "add_" + api_definition.name()
     required_params = api_definition.required_params
+    optional_params = api_definition.optional_params
     def fn(*args, **kwargs):
-        if len(args) != len(required_params):
-            raise TypeError("%s() takes exactly %d arguments (%d given)" \
+        if len(args) < len(required_params):
+            raise TypeError("%s() takes at least %d arguments (%d given)" \
                             % (fn_name, len(required_params), len(args)))
-        args = list(args)
-        for param in required_params:
-            kwargs[param] = args.pop(0)
+        max_args = len(required_params) + len(optional_params)
+        if len(args) > max_args:
+            raise TypeError("%s() takes at most %d arguments (%d given)" \
+                            % (fn_name, max_args, len(args)))
+        idx = -1
+        # Load kwargs with required params
+        for idx, param in enumerate(required_params):
+            kwargs[param] = args[idx]
+        idx += 1
+        # Load kwargs with unnamed optional params
+        for i in xrange(idx, len(args)):
+            kwargs[optional_params[i-idx]] = args[i]
+
         data = parent._post(api._uq_url, data=kwargs)
         ret = api(parent, data)
         return ret
@@ -245,7 +264,7 @@ def _add_api(definition, parent):
 
     if _LIST in definition.actions:
         list_fn = _add_list_fn(cls, definition, parent)
-        _add_find_fn(name, list_fn, parent)
+        _add_find_fn(cls, name, parent)
     if _GET in definition.actions:
         _add_get_fn(cls, name, parent)
     if _ADD in definition.actions:
@@ -277,6 +296,8 @@ class _GitLabAPI(object):
             setattr(self, '_id', json_data[self._key_name])
         except KeyError:  # some objects don't give us an id (e.g. events)
             pass
+        if self._convert_dates_enabled:
+            self._convert_dates(json_data)
         for key, val in json_data.iteritems():
             setattr(self, key, val)
         self._parent = parent
@@ -285,6 +306,51 @@ class _GitLabAPI(object):
             _add_api(sub_api, self)
         self._sub_apis = None
 
+    _date_fields = {
+        'created_at': True,
+        'updated_at': True,
+        'expires_at': True,
+        'last_activity_at': True,
+        'timestamp': True,
+        'authored_date': True,
+        'committed_date': True,
+    }
+    def _convert_dates(self, data):
+        if type(data) == list:
+            for item in data:
+                self._convert_dates(item)
+            return
+        for key, val in data.iteritems():
+            if type(val) == dict:
+                self._convert_dates(val)
+            if self._date_fields.get(key):
+                data[key] = self._convert_gitlab_date(val)
+
+    def _convert_gitlab_date(self, datetime_str):
+        """Convert GitLab datetime string to datetime object"""
+        fmt = '%Y-%m-%dT%H:%M:%S'
+        offset = None
+        if datetime_str.endswith('Z'):
+            datetime_str = datetime_str[:-1]
+        else:
+            offset = datetime_str[-6:]
+            datetime_str = datetime_str[:-6]
+        dt = datetime.strptime(datetime_str, fmt)
+        if not offset:
+            return dt
+        class GitLabTzInfo(tzinfo):
+            def __init__(self, utcoffset):
+                self.utcoffset_val = timedelta(minutes=utcoffset)
+            def utcoffset(self, dt):
+                return self.utcoffset_val
+        sign = offset[0]
+        hours = int(offset[1:3])
+        minutes = int(offset[-2:])
+        offset = hours*60 + minutes
+        if sign == '-':
+            offset = -offset
+        return dt.replace(tzinfo=GitLabTzInfo(offset))
+
     def _get_url(self, api_url, addl_keys=[]):
         keys = self._get_keys(addl_keys)
         # Handle annoying case of CurrentUser (wherein we have more keys
@@ -311,65 +377,48 @@ class _GitLabAPI(object):
         ret.reverse()  # Need to modify this later so no reversed()
         return ret
 
+    _code_to_exc = {
+        400: exceptions.MissingRequiredAttribute,
+        401: exceptions.UnauthorizedRequest,
+        403: exceptions.ForbiddenRequest,
+        404: exceptions.ResourceNotFound,
+        405: exceptions.RequestNotSupported,
+        409: exceptions.ResourceConflict,
+        500: exceptions.ServerError,
+    }
     def _check_status_code(self, status_code, url, data):
         if status_code < 400:
             return
         msg = "URL: %s, Data: %s" % (url, data)
-        if status_code == 400:
-            raise exceptions.MissingRequiredAttribute(msg)
-        elif status_code == 401:
-            raise exceptions.UnauthorizedRequest(msg)
-        elif status_code == 403:
-            raise exceptions.ForbiddenRequest(msg)
-        elif status_code == 404:
-            raise exceptions.ResourceNotFound(msg)
-        elif status_code == 405:
-            raise exceptions.RequestNotSupported(msg)
-        elif status_code == 409:
-            raise exceptions.ResourceConflict(msg)
-        elif status_code == 500:
-            raise exceptions.ServerError(msg)
+        raise self._code_to_exc[status_code](msg)
 
     def _get(self, api_url, addl_keys=[], data=None):
         """get or list"""
+        return self._request(requests.get, api_url, addl_keys, data)
+
+    def _post(self, api_url, addl_keys=[], data=None):
+        return self._request(requests.post, api_url, addl_keys, data)
+
+    def _put(self, api_url, addl_keys=[], data=None):
+        return self._request(requests.put, api_url, addl_keys, data)
+
+    def _delete(self, api_url, addl_keys=[], data=None):
+        return self._request(requests.delete, api_url, addl_keys, data)
+
+    def _request(self, request_fn, api_url, addl_keys, data):
         url = self._get_url(api_url, addl_keys)
+        #print "%s %s, data=%s" % (request_fn.__name__.upper(), url, str(data))
         try:
-            r = requests.get(url, headers=self._headers, data=data)
+            r = request_fn(url, headers=self._headers, data=data)
         except requests.exceptions.RequestException:
-            raise exceptions.ConnectionError("Failed to get " + url)
+            msg = "'%s' request to '%s' failed" % (request_fn.__name__, url)
+            raise exceptions.ConnectionError(msg)
         self._check_status_code(r.status_code, url, data)
         try:
             return json.loads(r.content)
         except ValueError:  # XXX: assume we're returning plain text
             return r.content
 
-    def _post(self, api_url, addl_keys=[], data=None):
-        url = self._get_url(api_url, addl_keys)
-        try:
-            r = requests.post(url, headers=self._headers, data=data)
-        except requests.exceptions.RequestException:
-            raise exceptions.ConnectionError("Failed to post " + url)
-        self._check_status_code(r.status_code, url, data)
-        return json.loads(r.content)
-
-    def _put(self, api_url, addl_keys=[], data=None):
-        url = self._get_url(api_url, addl_keys)
-        try:
-            r = requests.put(url, headers=self._headers, data=data)
-        except requests.exceptions.RequestException:
-            raise exceptions.ConnectionError("Failed to put " + url)
-        self._check_status_code(r.status_code, url, data)
-        return json.loads(r.content)
-
-    def _delete(self, api_url, addl_keys=[], data=None):
-        url = self._get_url(api_url, addl_keys)
-        try:
-            r = requests.delete(url, headers=self._headers, data=data)
-        except requests.exceptions.RequestException:
-            raise exceptions.ConnectionError("Failed to delete " + url)
-        self._check_status_code(r.status_code, url, data)
-        return json.loads(r.content)
-
     def __repr__(self):
         """__repr__ function for new API class"""
         return str(self._get_data())
@@ -378,9 +427,10 @@ class _GitLabAPI(object):
 class GitLab(_GitLabAPI):
     """A GitLab API connection."""
 
-    def __init__(self, gitlab_url, token=None):
+    def __init__(self, gitlab_url, token=None, convert_dates=True):
         setattr(_GitLabAPI, '_base_url', gitlab_url + "/api/v3")
         setattr(_GitLabAPI, '_headers', {'PRIVATE-TOKEN': token})
+        setattr(_GitLabAPI, '_convert_dates_enabled', convert_dates)
 
         for sub_api in _GitLabAPIDefinition.sub_apis:
             cls = _add_api(sub_api, self)
